#include "serial_port_worker.h"
#include "serial_port_thread.h"

#include <QDebug>
#include <QCoreApplication>

std::atomic<uint8_t> Command::m_nextSendBufferIndex{0};
std::atomic<uint8_t> Command::m_nextRecvBufferIndex{0};
char Command::m_sendBuffers[256][SEND_BLOCK_SIZE]{{0}};
char Command::m_recvBuffers[256][SEND_BLOCK_SIZE]{{0}};

uint32_t Command::getNextCommandPriority(uint8_t priorityClass) {
    static std::atomic<std::uint32_t> priority_1{0};
    static std::atomic<std::uint32_t> priority_2{1000000000};
    static std::atomic<std::uint32_t> priority_3{2000000000};
    static std::atomic<std::uint32_t> priority_4{3000000000};
    switch (priorityClass) {
    case PRIORITY_1:
        return priority_1.fetch_add(1);
    case PRIORITY_2:
        return priority_2.fetch_add(1);
    case PRIORITY_3:
        return priority_3.fetch_add(1);
    case PRIORITY_4:
        return priority_4.fetch_add(1);
    }
    return -1;
}

char *Command::getNextSendBuffer() {
    uint8_t const nextIndex = m_nextSendBufferIndex.fetch_add(1);
    return &m_sendBuffers[nextIndex][0];
}

char *Command::getNextRecvBuffer() {
    uint8_t const nextIndex = m_nextRecvBufferIndex.fetch_add(1);
    return &m_recvBuffers[nextIndex][0];
}

Command::Command(uint16_t cmdId, uint8_t sendBufferLength,
                 uint8_t recvBufferLength, uint8_t priorityClass)
    : m_priority(getNextCommandPriority(priorityClass))
    , m_priorityClass(priorityClass)
    , m_cmdId(cmdId)
    , m_sendBuffer(Command::getNextSendBuffer())
    , m_recvBuffer(Command::getNextRecvBuffer())
    , m_sendBufferLength(sendBufferLength)
    , m_recvBufferLength(recvBufferLength) {
    memset(m_sendBuffer, 0x00, sizeof(m_sendBuffers[0]));
}

Command::Command()
    : m_priority(0)
    , m_priorityClass(0)
    , m_cmdId(0)
    , m_sendBuffer(0)
    , m_recvBuffer(0)
    , m_sendBufferLength(0)
    , m_recvBufferLength(0) {
}

/**
* \brief Constructor for SerialPort
*
* @date 08.11.2022 – first draft.
*
*/
SerialPortWorker::SerialPortWorker(QString portName, enum QSerialPort::BaudRate baudRate, QObject *parent)
    : QThread(parent)
    , m_serialPort()
    , m_serialPortOpened(false)
    , m_portName(portName)
    , m_baudRate(baudRate)
    , m_bytesWritten(0) {
    this->setObjectName("SerialPortWorker");

    foreach (const QSerialPortInfo &info, QSerialPortInfo::availablePorts()) {
        if (!info.isNull()) {
            qDebug() << "Name .......... " << info.portName();
            qDebug() << "Description ... " << info.description();
            qDebug() << "Manufacturer .. " << info.manufacturer();
            qDebug() << "SerialNumber .. " << info.serialNumber();
            if (info.hasProductIdentifier()) {
                qDebug() << "Product-ID .... " << info.productIdentifier();
            }
            if (info.hasVendorIdentifier()) {
                qDebug() << "Vendor-ID ..... " << info.vendorIdentifier();
            }
            qDebug() << "Location ...... " << info.systemLocation();

            if (info.portName() == m_portName) {
                m_serialPortInfo = info;
            }
        }
    }

    if (m_serialPortInfo.portName() == m_portName) {
        m_serialPort.setPort(m_serialPortInfo);
        m_serialPort.setPortName(m_portName);
        m_serialPort.setDataBits(QSerialPort::Data8);
        m_serialPort.setParity(QSerialPort::NoParity);
        m_serialPort.setStopBits(QSerialPort::OneStop);
        m_serialPort.setFlowControl(QSerialPort::NoFlowControl);
        m_serialPortOpened = m_serialPort.open(QIODevice::ReadWrite);
    }
}

SerialPortWorker::~SerialPortWorker() {
    closeSerialPort();
}

void SerialPortWorker::run() {
    SerialPortThread serialPortThread;
    this->moveToThread(&serialPortThread); // not really necessary

    // Note: because this thread is moved to serialPortThread, the event loop
    // used will be the one of serialPortThread (this-thread does not even have one).
    // This is the only purpose of serialPortThread.
    serialPortThread.start();

    connect(&m_serialPort, SIGNAL(readyRead()), this, SLOT(getBytesAvailable()));
    connect(&m_serialPort, SIGNAL(bytesWritten(qint64)), this, SLOT(getNumberOfBytesWritten(qint64)));

    while (m_serialPortOpened) {
        Command nextCmd = getNextCommand();
        if (nextCmd.m_cmdId == Command::QUIT_SENTINEL) {
            break;
        }
        QByteArray data(nextCmd.getNextSendBuffer(), nextCmd.m_sendBufferLength);
        m_bytesWritten = 0;
        m_serialPort.write(data);
        // process events in event loop of associated thread
        serialPortThread.getEventLoop().processEvents(QEventLoop::ExcludeUserInputEvents, TIMEOUT);
        int maxWaitTime = WAIT_TIME_MAX;

        while (maxWaitTime > 0 && m_bytesWritten < (uint32_t)data.size()) {
            if (!m_serialPort.waitForBytesWritten(100)) { // wait timeout-ms for the bytes to be written
                // process events in event loop of associated thread
                serialPortThread.getEventLoop().processEvents(QEventLoop::ExcludeUserInputEvents, TIMEOUT);
                maxWaitTime -= TIMEOUT;
            }
        }
        if (m_bytesWritten == (uint32_t)data.size()) { // data written, wait for result
            if (nextCmd.m_recvBufferLength > 0) {
                // process events in event loop of associated thread
                serialPortThread.getEventLoop().processEvents(QEventLoop::ExcludeUserInputEvents, TIMEOUT);
                maxWaitTime = WAIT_TIME_MAX;
                while (maxWaitTime > 0 && m_bytesAvailable < nextCmd.m_recvBufferLength) {
                    if (!m_serialPort.waitForReadyRead(100)) {
                        // process events in event loop of associated thread
                        serialPortThread.getEventLoop().processEvents(QEventLoop::ExcludeUserInputEvents, TIMEOUT);
                        maxWaitTime -= TIMEOUT;
                    }
                }
                uint32_t const bytesAvailable = m_bytesAvailable;
                if (bytesAvailable == nextCmd.m_recvBufferLength) {
                    QByteArray result = m_serialPort.read(bytesAvailable);
                    if (!result.isEmpty() && ((uint32_t)result.size() == bytesAvailable)) {
                        memcpy(nextCmd.m_recvBuffer, result.data(), bytesAvailable);
                        insertResult(nextCmd);
                    }
                } else {
                    // error
                }
            }
        } else {
            // error
        }
    } // while (m_serialPortOpened)

    serialPortThread.quit();
    serialPortThread.wait();
}

bool SerialPortWorker::openSerialPort() {
    if (!m_serialPortOpened) {
        return m_serialPort.open(QIODevice::ReadWrite);
    }
    return false; // opening twice is not allowed
}

void SerialPortWorker::closeSerialPort() {
    if (m_serialPortOpened) {
        m_serialPort.close();
        m_serialPortOpened = false;
    }
}

void SerialPortWorker::quit() {
    Command cmd(Command::QUIT_SENTINEL, 0, 0, Command::PRIORITY_1);
    insertCommand(cmd);
}

void SerialPortWorker::insertCommand(Command const &cmd) {
    {   std::lock_guard<std::mutex> lock(m_sendQueueMutex);
        m_sendQueue.push(cmd);
    } // release lock
    m_sendQueueCondVar.notify_one();
}

Command SerialPortWorker::getNextCommand() {
    Command nextCmd;
    {   std::unique_lock<std::mutex> lock(m_sendQueueMutex);
        m_sendQueueCondVar.wait(lock, [this]{ return !this->m_sendQueue.empty(); });
        nextCmd = m_sendQueue.top();
        m_sendQueue.pop();
    } // release lock
    return nextCmd;
}

void SerialPortWorker::insertResult(Command const &result) {
    {   std::lock_guard<std::mutex> lock(m_recvQueueMutex);
        m_recvQueue.push_back(result);
    } // release lock
    m_recvQueueCondVar.notify_one();
}

bool SerialPortWorker::getNextResult(Command &nextResult, int timeout) {
    bool timed_out = false;
    {   std::unique_lock<std::mutex> lock(m_recvQueueMutex);
        if (m_recvQueueCondVar.wait_for(lock, std::chrono::milliseconds(timeout),
                                    [this]{ return !this->m_recvQueue.empty(); })) {
            nextResult = m_recvQueue.front();
            m_recvQueue.pop_front();
            return true;
        } else {
            timed_out = true;
        }
    } // release lock
    return (timed_out == false);
}

void SerialPortWorker::getBytesAvailable() {
    m_bytesAvailable += m_serialPort.bytesAvailable();
}

bool SerialPortWorker::getNumberOfBytesWritten(qint64 bytesWritten) {
    if (bytesWritten > 0) {
        m_bytesWritten += (uint32_t)bytesWritten;
        return true;
    }
    return false;
}

